/* eslint-disable max-len */
const postcss = require('postcss')
const valueParser = require('postcss-value-parser')

module.exports = postcss.plugin('postcss-css-variable-value', options => {
  options = options || { preserve: true }

  // is this a CSS variable
  function isCssVarClause (value) {
    // if it DOES contain `var(` it is a CSS variable clause
    return !!value.includes('var(')
  }

  // is this a CSS variable reassignment
  function isCssVarReassignment (value) {
    // if it DOES NOT contain a comma it is a reassignment
    return !value.includes(',')
  }

  // recursive drill in to CSS variable
  function findFallbackRecursively (value) {
    // get the fallback value from the CSS variable clause
    // regex:
    //    var\(                 -> find something that starts with `var(`
    //    (?=\S*['-]*)          -> positive lookahead, zero or more non-space characters and zero or more ' or - characters
    //    ([a-zA-Z'-]+)         -> match group 1, one or more alpha characters
    //    \s*,\s*               -> zero or more spaces, a comma, zero or more spaces
    //    (.+)                  -> match group 2, one or more of any character
    //    \)                    -> a closing parenthesis
    let regex = /var\((?=\S*['-]*)([a-zA-Z'-]+)\s*,\s*(.+)\)/
    let matches = value.match(regex)

    if (matches) {
      let fallbackValue = matches[2]

      // if we have another CSS variable clause then keep drilling in
      if (isCssVarClause(fallbackValue)) {
        return findFallbackRecursively(fallbackValue)
      } else {
        return matches[2].trim()
      }
    }

    // if we have a reassignment
    if (isCssVarReassignment(value)) {
      return null
    }

    return value
  }

  function parseDelimited (delimitedValue) {
    let tokenArgParsedValue = delimitedValue
    let TOKEN_DELIMITER = /\|\|\[\[(\w+)\]\]\|\|/
    let replacement

    // check if the first node type = word and begins and ends with the delimiter
    // then conditionally swap out the delimiter
    if (RegExp(`^${ TOKEN_DELIMITER.source }$`).test(tokenArgParsedValue)) {
      replacement = '$1'
    } else {
      replacement = "'+$1+'"
      tokenArgParsedValue = `'${ tokenArgParsedValue }'`
    }

    return tokenArgParsedValue.replace(TOKEN_DELIMITER, replacement)
  }

  // parse token(...) containing a var(...) function to a simple form
  function parseTokenFunction (nodes) {
    let nodesValue = nodes

    let tokenArgParsable = valueParser
      .stringify(nodesValue)
      .replace(/'\+(.+)\+'/g, '||[[$1]]||')
      .replace(/'(.+)'/g, '$1')

    let parsedValue = valueParser(tokenArgParsable)

    let tokenArgNodeArray = []
    parsedValue.walk(tokenArgNode => {
      tokenArgNodeArray.push(tokenArgNode.type)
    })

    if (!tokenArgNodeArray.includes('function')) return false

    parsedValue.walk(tokenArgNode => {
      if (tokenArgNode.type === 'function' && tokenArgNode.value === 'var') {
        // this converts function nodes of type = var to word nodes containing the fallback
        tokenArgNode.value = findFallbackRecursively(
          valueParser.stringify(tokenArgNode)
        )
        tokenArgNode.type = 'word'
      }
    })

    if (parsedValue.toString() === tokenArgParsable) return false

    nodesValue = [
      { type: 'word', value: parseDelimited(parsedValue.toString()) }
    ]

    return nodesValue
  }

  return css => {
    css.walkRules(rule => {
      rule.walkDecls(decl => {
        let hasChanges = false
        let declaration = decl

        // if we'd like to preserve the existing declaration as is then keep it and process a clone instead
        if (options.preserve) {
          declaration = decl.clone()
        }

        let parsedValue = valueParser(declaration.value)

        parsedValue.walk(node => {
          // Since we only want to transform var(...) function values,
          // we can ignore anything else.
          if (node.type !== 'function') return
          if (node.value !== 'var' && node.value !== 'token') return

          if (node.value === 'var') {
            let numParams
            // calculate how many params the value has
            if (node.nodes) {
              numParams = node.nodes.filter(n => {
                return ['word', 'function', 'string'].some(v =>
                  n.type.includes(v)
                )
              }).length
            }

            // if we have less than 2 we don't have a fallback so do nothing
            if (numParams < 2) return
          }

          // parse token(...) functions that contain a var(...) function
          if (node.value === 'token') {
            let parsedTokenArgNodes = parseTokenFunction(node.nodes)
            if (parsedTokenArgNodes) {
              node.nodes = parsedTokenArgNodes
            } else {
              return
            }
          } else {
            // get the string value for this function
            let nodeValue = valueParser.stringify(node)

            // set the node type to generic word
            node.type = 'word'
            // parse down to the fallback value
            node.value = findFallbackRecursively(nodeValue)

            if (node.value === null) {
              return
            }
          }

          hasChanges = true
        })

        // set the declaration's value to the newly parsed one
        if (hasChanges) {
          declaration.value = parsedValue
        } else {
          declaration = null
        }

        // if there were changes and we'd like to preserve the original declaration then insert the new declaration before the existing one
        if (options.preserve && hasChanges) {
          decl.before(declaration)
        }
      })
    })
  }
})
